using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust
{

    /// <summary>
    /// Some basic help functionallity for lights.
    /// </summary>
    public class QuickLightning : MVRScript
    {
        private JSONStorableBool quickLightningEnabled;
        
        private JSONStorableBool strobeEnabled;
        private JSONStorableFloat strobeInterval;
        private JSONStorableStringChooser strobeType;

        private JSONStorableBool colorTweenEnabled;
        private JSONStorableFloat colorTweenSpeed;
        private JSONStorableStringChooser colorTweenType;

        private JSONStorableFloat baseIntensity;
        private JSONStorableColor baseColor;

        private JSONStorableFloat flashIntensity;
        private JSONStorableFloat flashDuration;

        private JSONStorableColor colorTweenA, colorTweenB, colorTweenC;
        private JSONStorableBool colorTweenWrapHue;

        private JSONStorableFloat previewCurrentIntensity;
        private JSONStorableColor previewCurrentColor;
        
        private float m_tweenCurrentTime = 0f;
        private float m_strobeCurrentTime = 0f;

        private bool m_fadingIn = false;
        private bool m_fadingOut = false;
        private bool m_flashing = false;
        private float m_flashCurrentTime = 0f;

        public override void Init()
        {
            JSONStorable light = GetLightStorable(containingAtom);

            // Some debugging:
            // SuperController.LogMessage(string.Join(", ", containingAtom.GetStorableIDs().ToArray()));
            // if (light != null)
            // {
                // SuperController.LogMessage("GetBoolParamNames: " + string.Join(", ", light.GetBoolParamNames().ToArray())); // on, (showDust), (showHalo), shadowsOn
                // SuperController.LogMessage("GetFloatParamNames: " + string.Join(", ", light.GetFloatParamNames().ToArray())); // intensity, range, spotAngle, pointBias
                // SuperController.LogMessage("GetVector3ParamNames: " + string.Join(", ", light.GetVector3ParamNames().ToArray()));
                // SuperController.LogMessage("GetStringParamNames: " + string.Join(", ", light.GetStringParamNames().ToArray()));
                // SuperController.LogMessage("GetColorParamNames: " + string.Join(", ", light.GetColorParamNames().ToArray())); // color
                // SuperController.LogMessage("GetActionNames: " + string.Join(", ", light.GetActionNames().ToArray()));
                // SuperController.LogMessage("GetAudioClipActionNames: " + string.Join(", ", light.GetAudioClipActionNames().ToArray()));
                // SuperController.LogMessage("GetUrlParamNames: " + string.Join(", ", light.GetUrlParamNames().ToArray()));
                // SuperController.LogMessage("GetAllParamAndActionNames: " + string.Join(", ", light.GetAllParamAndActionNames().ToArray()));
            // }

            bool lightEnabledDefaultValue = true;
            if (light != null)
            {
                lightEnabledDefaultValue = light.GetBoolParamValue("on");
            }
            // Boolean for option Enabled (plugin enabled and running, if Strobe is true the light will blink even if saved in "off" state)
            quickLightningEnabled = new JSONStorableBool("LightEnabled", lightEnabledDefaultValue, (bool value) => {
                try
                {
                    if (value)
                    {
                        TurnOn();
                    }
                    else
                    {
                        TurnOff();
                    }
                } 
                catch (Exception e)
                {
                    SuperController.LogError("Exception in LightEnabled: " + e.ToString());
                }
            });
            RegisterBool(quickLightningEnabled);
            UIDynamicToggle toggle = CreateToggle(quickLightningEnabled, false);
            toggle.label = "LightEnabled";

            // Register action callbacks to be used for scripting and scene creation
            CreateAction("TurnLightOn", () => { TurnOn(); }, false);
            CreateAction("TurnLightOff", () => { TurnOff(); }, false);
            CreateAction("ToggleLightOnOff", () => {
                if (quickLightningEnabled.val)
                {
                    TurnOff();
                }
                else
                {
                    TurnOn();
                }
            }, false);

            // TurnLightOffFadeOut : Fades out the light and turns if off.
            CreateAction("TurnLightOnFadeIn", () => { TurnOnFadeIn(); }, false);
            CreateAction("TurnLightOffFadeOut", () => { TurnOffFadeOut(); }, false);

            // TurnOffAllLightsInScene : Turns off all lights in the scene, used for lightning triggers to reset the scene easily.
            CreateAction("TurnOffAllLightsInScene", () => { TurnOffAll(false); }, false);

            // TurnOffFadeOutAllLightsInScene : Fades out all lights in the scene that has this plugin attached (otherwise it turns the light off), used for lightning triggers to reset the scene easily or create effects.
            CreateAction("TurnOffFadeOutAllLightsInScene", () => { TurnOffAll(true); }, false);

            // Save and Restore state is only within the current scene session (is not saved together with the scene)
            CreateAction("SaveState", () => { SaveState(); }, false);
            CreateAction("RestoreState", () => { RestoreState(); }, false);

            CreateAction("SaveStateAllLightsInScene", () => { SaveStateAll(); }, false);
            CreateAction("RestoreStateAllLightsInScene", () => { RestoreStateAll(); }, false);

            CreateSpacer(false);

            // Boolean for option Strobe
            strobeEnabled = new JSONStorableBool("strobeEnabled", false, (bool value) => { 
                if (value)
                {
                    m_strobeCurrentTime = 0f;
                } 
                else
                {
                    RestoreBaseIntensity();
                }
            });
            RegisterBool(strobeEnabled);
            toggle = CreateToggle(strobeEnabled, false);
            toggle.label = "strobeEnabled";

            List<string> strobeTypes = new List<string>();
            strobeTypes.Add("Fixed");
            strobeTypes.Add("Lerp");
            strobeTypes.Add("Smooth");
            strobeType = new JSONStorableStringChooser("strobeType", strobeTypes, "Fixed", "strobeType", (string type) => { });
            RegisterStringChooser(strobeType);
            CreateScrollablePopup(strobeType, false);

            // float param for time interval between on/off during strobe
            strobeInterval = new JSONStorableFloat("strobeInterval(ms)", 50.0f, (JSONStorableFloat jf) => { }, 10f, 1000f, false);
            RegisterFloat(strobeInterval);
            UIDynamicSlider slider = CreateSlider(strobeInterval, false);
            slider.valueFormat = "F1";
            slider.ConfigureQuickButtons(-1f, -10f, -50f, -100f, 1f, 10f, 50f, 100f);

            // Actions for strobe control
            CreateAction("SetStrobeOn", () => { strobeEnabled.val = true; }, false);
            CreateAction("SetStrobeOff", () => { strobeEnabled.val = false; }, false);
            CreateAction("ToggleStrobeOnOff", () => { strobeEnabled.val = !strobeEnabled.val; }, false);

            // Flash : Flashes with the light and turns if off
            CreateAction("Flash", () => { Flash(); }, true);

            // slider for flash intensity
            flashIntensity = new JSONStorableFloat("flashIntensity", 6.0f, (JSONStorableFloat jf) => { }, 0f, 8f, true);
            RegisterFloat(flashIntensity);
            slider = CreateSlider(flashIntensity, false);
            slider.valueFormat = "F3";
            slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

            // slider for flash duration
            flashDuration = new JSONStorableFloat("flashDuration(ms)", 300.0f, (JSONStorableFloat jf) => { }, 0f, 1000f, false);
            RegisterFloat(flashDuration);
            slider = CreateSlider(flashDuration, false);
            slider.valueFormat = "F1";
            slider.ConfigureQuickButtons(-1f, -10f, -50f, -100f, 1f, 10f, 50f, 100f);

            CreateSpacer(false);

            // Boolean for option ColorTweening
            colorTweenEnabled = new JSONStorableBool("colorTweenEnabled", false, (bool value) => { 
                if (value)
                {
                    m_tweenCurrentTime = 0f;
                } 
                else
                {
                    RestoreBaseColor();
                }
            });
            RegisterBool(colorTweenEnabled);
            toggle = CreateToggle(colorTweenEnabled, false);
            toggle.label = "colorTweenEnabled";

            List<string> tweenTypes = new List<string>();
            tweenTypes.Add("Fixed");
            tweenTypes.Add("Lerp");
            tweenTypes.Add("Smooth");
            colorTweenType = new JSONStorableStringChooser("colorTweenType", tweenTypes, "Lerp", "colorTweenType", (string type) => { });
            RegisterStringChooser(colorTweenType);
            CreateScrollablePopup(colorTweenType, false);

            // float param for speed for ColorTweening
            colorTweenSpeed = new JSONStorableFloat("colorTweenSpeed(ms)", 300.0f, (JSONStorableFloat jf) => { }, 10f, 1000f, false);
            RegisterFloat(colorTweenSpeed);
            slider = CreateSlider(colorTweenSpeed, false);
            slider.valueFormat = "F1";
            slider.ConfigureQuickButtons(-1f, -10f, -50f, -100f, 1f, 10f, 50f, 100f);

            colorTweenWrapHue = new JSONStorableBool("colorTweenWrapHue", true, (bool value) => { });
            RegisterBool(colorTweenWrapHue);
            toggle = CreateToggle(colorTweenWrapHue, false);
            toggle.label = "WrapHue";

            CreateAction("SetColorTweenOn", () => { colorTweenEnabled.val = true; }, false);
            CreateAction("SetColorTweenOff", () => { colorTweenEnabled.val = false; }, false);
            CreateAction("ToggleColorTweenOnOff", () => { colorTweenEnabled.val = !colorTweenEnabled.val; }, false);
            
            CreateSpacer(false);

            // Intensity slider
            baseIntensity = new JSONStorableFloat("baseIntensity", 1f, (JSONStorableFloat jf) => {
                light = GetLightStorable(containingAtom);
                if (light != null)
                {
                    light.SetFloatParamValue("intensity", jf.val);
                }
            }, 0f, 8f, true);
            RegisterFloat(baseIntensity);
            slider = CreateSlider(baseIntensity, true);
            slider.valueFormat = "F3";
            slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 1f, 0.1f, 0.01f, 0.001f);

            // Color Pickers
            HSVColor c = new HSVColor();
            c.H = 0.1f;
            c.S = 1f;
            c.V = 0.8f;
            baseColor = new JSONStorableColor("baseColor", c, (JSONStorableColor js) => { RestoreBaseColor(); });
            RegisterColor(baseColor);
            CreateColorPicker(baseColor, true);

            // 3 color pickers for tweening color
            c = new HSVColor();
            c.H = 0.8f;
            c.S = 0.75f;
            c.V = 1.0f;
            colorTweenA = new JSONStorableColor("tweenColorA", c);
            RegisterColor(colorTweenA);
            CreateColorPicker(colorTweenA, true);

            c = new HSVColor();
            c.H = 0.5f;
            c.S = 1.0f;
            c.V = 1.0f;
            colorTweenB = new JSONStorableColor("tweenColorB", c);
            RegisterColor(colorTweenB);
            CreateColorPicker(colorTweenB, true);

            c = new HSVColor();
            c.H = 0.15f;
            c.S = 0.5f;
            c.V = 1.0f;
            colorTweenC = new JSONStorableColor("tweenColorC", c);
            RegisterColor(colorTweenC);
            CreateColorPicker(colorTweenC, true);

            // preview intensity slider (readonly)
            float intensityStart = 1f;
            if (light != null)
            {
                intensityStart = light.GetFloatParamValue("intensity");
            }
            previewCurrentIntensity = new JSONStorableFloat("previewCurrentIntensity", intensityStart, (JSONStorableFloat jf) => { }, 0f, 8f, true);
            CreateSlider(previewCurrentIntensity, true);

            // preview color picker (readonly) for current color.
            c = new HSVColor();
            c.H = 0f;
            c.S = 1f;
            c.V = 1f;
            if (light != null)
            {
                c = light.GetColorParamValue("color");
            }
            previewCurrentColor = new JSONStorableColor("previewCurrentColor", c);
            CreateColorPicker(previewCurrentColor, true);
        }

        protected void Update()
        {
            JSONStorable light = GetLightStorable(containingAtom);
            if (light != null)
            {
                bool on = light.GetBoolParamValue("on");
                if (quickLightningEnabled.val != on)
                {
                    quickLightningEnabled.val = on;
                }
            }

            if (!quickLightningEnabled.val)
            {
                return;
            }

            if (colorTweenEnabled.val || strobeEnabled.val || m_flashing || m_fadingOut || m_fadingIn)
            {
                if (colorTweenEnabled.val)
                {
                    HSVColor prevC = new HSVColor();
                    HSVColor targetC = new HSVColor();
                    float totTween = colorTweenSpeed.val * 3f;
                    float p = 1f;
                    m_tweenCurrentTime += Time.deltaTime * 1000f;
                    if (m_tweenCurrentTime >= totTween)
                    {
                        m_tweenCurrentTime -= (int)(m_tweenCurrentTime / totTween) * totTween;
                    }

                    if (m_tweenCurrentTime <= colorTweenSpeed.val)
                    {
                        prevC = colorTweenC.val;
                        targetC = colorTweenA.val;
                        p = m_tweenCurrentTime / colorTweenSpeed.val;
                    }
                    else if (m_tweenCurrentTime <= colorTweenSpeed.val * 2f)
                    {
                        prevC = colorTweenA.val;
                        targetC = colorTweenB.val;
                        p = (m_tweenCurrentTime - colorTweenSpeed.val) / colorTweenSpeed.val;
                    }
                    else
                    {
                        // From third to first
                        prevC = colorTweenB.val;
                        targetC = colorTweenC.val;
                        p = (m_tweenCurrentTime - colorTweenSpeed.val * 2f) / colorTweenSpeed.val;
                    }

                    switch (colorTweenType.val)
                    {
                        case "Fixed": break;
                        case "Lerp":
                            {
                                float dist = Math.Abs(targetC.H - prevC.H);
                                float wrapDist = 1f - dist;

                                if (dist < wrapDist || !colorTweenWrapHue.val)
                                {
                                    targetC.H = Mathf.Lerp(prevC.H, targetC.H, p);
                                }
                                else
                                {
                                    // increase and wrap-around or decrease and wrap-around?
                                    if (prevC.H < targetC.H)
                                    {
                                        // decrease
                                        float h = Mathf.Lerp(prevC.H, targetC.H - 1f, p);
                                        if (h < 0f)
                                            h += 1f;

                                        targetC.H = h;
                                    }
                                    else
                                    {
                                        // increase
                                        float h = Mathf.Lerp(prevC.H, targetC.H + 1f, p);
                                        if (h > 1f)
                                            h -= 1f;
                                        targetC.H = h;
                                    }
                                }

                                // targetC.H = Mathf.Lerp(prevC.H, targetC.H, p);
                                targetC.S = Mathf.Lerp(prevC.S, targetC.S, p);
                                targetC.V = Mathf.Lerp(prevC.V, targetC.V, p);
                            }
                            break;
                        case "Smooth":
                            {
                                float dist = Math.Abs(targetC.H - prevC.H);
                                float wrapDist = 1f - dist;

                                if (dist < wrapDist || !colorTweenWrapHue.val)
                                {
                                    targetC.H = Mathf.Lerp(prevC.H, targetC.H, p);
                                }
                                else
                                {
                                    // increase and wrap-around or decrease and wrap-around?
                                    if (prevC.H < targetC.H)
                                    {
                                        // decrease
                                        float h = Mathf.SmoothStep(prevC.H, targetC.H - 1f, p);
                                        if (h < 0f)
                                            h += 1f;

                                        targetC.H = h;
                                    }
                                    else
                                    {
                                        // increase
                                        float h = Mathf.SmoothStep(prevC.H, targetC.H + 1f, p);
                                        if (h > 1f)
                                            h -= 1f;
                                        targetC.H = h;
                                    }
                                }

                                // targetC.H = Mathf.SmoothStep(prevC.H, targetC.H, p);
                                targetC.S = Mathf.SmoothStep(prevC.S, targetC.S, p);
                                targetC.V = Mathf.SmoothStep(prevC.V, targetC.V, p);
                            }
                            break;
                    }

                    if (light != null)
                    {
                        light.SetColorParamValue("color", targetC);
                    }
                    previewCurrentColor.val = targetC;
                }

                if (m_fadingIn)
                {
                    // NOTE: FadeOut also uses m_flashCurrentTime and FlashDuration
                    m_flashCurrentTime += Time.deltaTime * 1000f;
                    float p = Math.Min((m_flashCurrentTime / flashDuration.val), 1f);

                    float targetIntensity = baseIntensity.val * p;
                    targetIntensity = Math.Max(0f, targetIntensity);

                    if (light != null)
                    {
                        light.SetFloatParamValue("intensity", targetIntensity);
                    }
                    previewCurrentIntensity.val = targetIntensity;

                    if (m_flashCurrentTime >= flashDuration.val)
                    {
                        m_fadingIn = false;
                    }
                } 
                else if (m_fadingOut)
                {
                    // NOTE: FadeOut also uses m_flashCurrentTime and FlashDuration
                    m_flashCurrentTime += Time.deltaTime * 1000f;
                    float p = 1f - (m_flashCurrentTime / flashDuration.val);

                    float targetIntensity = baseIntensity.val * p;
                    targetIntensity = Math.Max(0f, targetIntensity);

                    if (light != null)
                    {
                        light.SetFloatParamValue("intensity", targetIntensity);
                    }
                    previewCurrentIntensity.val = targetIntensity;

                    if (m_flashCurrentTime >= flashDuration.val)
                    {
                        m_fadingOut = false;
                        TurnOff();
                    }
                } 
                else if (m_flashing)
                {
                    m_flashCurrentTime += Time.deltaTime * 1000f;
                    float p = m_flashCurrentTime / flashDuration.val;
                    // first 0-20% of section will be increased intensity
                    // first 20-30% of section will be peak intensity
                    // first 30-100% of section will be back to 0 intensity
                    const float A = 0.2f;
                    const float B = 0.3f;

                    float targetIntensity = flashIntensity.val;
                    if (p <= A)
                    {
                        p = (p / A);
                    } 
                    else if (p <= B)
                    {
                        p = 1f;
                    } 
                    else
                    {
                        p = 1f - ((p - B) / (1f - B));
                    } 
                    
                    targetIntensity = Math.Max(0f, p * Math.Max(0f, targetIntensity));

                    if (light != null)
                    {
                        light.SetFloatParamValue("intensity", targetIntensity);
                    }
                    previewCurrentIntensity.val = targetIntensity;

                    if (m_flashCurrentTime >= flashDuration.val)
                    {
                        m_flashing = false;
                        TurnOff();
                    }
                    
                }
                else if (strobeEnabled.val)
                {
                    float totStrobe = strobeInterval.val * 2f;
                    float p = 1f;
                    float targetIntensity = baseIntensity.val;
                    m_strobeCurrentTime += Time.deltaTime * 1000f;
                    if (m_strobeCurrentTime >= totStrobe)
                    {
                        m_strobeCurrentTime -= (int)(m_strobeCurrentTime / totStrobe) * totStrobe;
                    }
                    if (m_strobeCurrentTime <= strobeInterval.val)
                    {
                        // first cycle
                        p = 1f - (m_strobeCurrentTime / strobeInterval.val);
                    } 
                    else
                    {
                        // second cycle
                        p = (m_strobeCurrentTime - strobeInterval.val) / strobeInterval.val;
                    }

                    switch(strobeType.val)
                    {
                        case "Fixed":
                            if (p < 0.5f)
                            {
                                targetIntensity = 0f;
                            }
                            break;
                        case "Lerp":
                            targetIntensity = Mathf.Lerp(0f, targetIntensity, p);
                            break;
                        case "Smooth":
                            targetIntensity = Mathf.SmoothStep(0f, targetIntensity, p);
                            break;
                    }
                    targetIntensity = Math.Max(0f, targetIntensity);

                    if (light != null)
                    {
                        light.SetFloatParamValue("intensity", targetIntensity);
                    }
                    previewCurrentIntensity.val = targetIntensity;
                }
            }
        }

        private JSONStorableAction CreateAction(string name, Action callback, bool createButton = true)
        {
            JSONStorableAction action = new JSONStorableAction(name, () => {
                try
                {
                    callback.Invoke();
                }
                catch (Exception e)
                {
                    SuperController.LogError("Exception in " + name + ": " + e.ToString());
                }
            });
            RegisterAction(action);
            if (createButton)
            {
                CreateButton(name, false).button.onClick.AddListener(() =>
                {
                    try
                    {
                        callback.Invoke();
                    }
                    catch (Exception e)
                    {
                        SuperController.LogError("Exception in " + name + ": " + e.ToString());
                    }
                });
            }
            return action;
        }

        private static JSONStorable GetLightStorable(Atom atom)
        {
            JSONStorable light = atom.GetStorableByID("Light");
            if (light == null)
                light = atom.GetStorableByID("TechnoRingLightLightSource");
            if (light == null)
                light = atom.GetStorableByID("TechnoNeonTriangleLightSource");
            if (light == null)
                light = atom.GetStorableByID("TechnoNeonSquareLightSource");
            if (light == null)
                light = atom.GetStorableByID("TechnoNeonHeartLightSource");
            if (light == null)
                light = atom.GetStorableByID("TechnoNeonCircleLightSource");
            if (light == null)
                light = atom.GetStorableByID("TechnoLightBarLightSource");
            return light;
        }

        private void TurnOnFadeIn()
        {
            if (!quickLightningEnabled.val || m_fadingOut)
            {
                TurnOn();
                JSONStorable light = GetLightStorable(containingAtom);
                if (light != null)
                {
                    light.SetFloatParamValue("intensity", 0f);
                    previewCurrentIntensity.val = 0f;
                }
                m_fadingIn = true;
                m_fadingOut = false;
                // NOTE: FadeOut also uses m_flashCurrentTime and FlashDuration
                m_flashCurrentTime = 0f;
            }
        }

        private void TurnOffFadeOut()
        {
            if (quickLightningEnabled.val || m_fadingIn)
            {
                m_fadingIn = false;
                m_fadingOut = true;
                // NOTE: FadeOut also uses m_flashCurrentTime and FlashDuration
                m_flashCurrentTime = 0f;
            }
        }


        private void Flash()
        {
            TurnOn();
            strobeEnabled.val = false;
            m_flashCurrentTime = 0f;
            m_flashing = true;
            m_fadingOut = false;
            m_fadingIn = false;
        }

        private void TurnOn()
        {
            JSONStorable light = GetLightStorable(containingAtom);
            if (light != null)
            {
                light.SetBoolParamValue("on", true);
            }
            if (!quickLightningEnabled.val)
            {
                quickLightningEnabled.val = true;
            }
            m_flashing = false;
            m_fadingOut = false;
            m_fadingIn = false;
            RestoreBaseIntensity();
            RestoreBaseColor();
        }

        private void TurnOff()
        {
            JSONStorable light = GetLightStorable(containingAtom);
            if (light != null)
            {
                light.SetBoolParamValue("on", false);
            }
            if (quickLightningEnabled.val)
            {
                quickLightningEnabled.val = false;
            }
            m_flashing = false;
            m_fadingOut = false;
            m_fadingIn = false;
        }

        private void TurnOffAll(bool fadeOut)
        {
            if (fadeOut)
            {
                TurnOffFadeOut();
                CallActionAllQuickLightning("TurnLightOffFadeOut", true);
            }
            else
            {
                TurnOff();
                CallActionAllQuickLightning("TurnLightOff", true);
            }
        }

        private bool store_quickLightningEnabled = true;
        private bool store_strobeEnabled = false;
        private bool store_colorTweenEnabled = false;

        private void SaveState()
        {
            // remember current primary states
            store_quickLightningEnabled = quickLightningEnabled.val;
            store_strobeEnabled = strobeEnabled.val;
            store_colorTweenEnabled = colorTweenEnabled.val;
        }

        private void RestoreState()
        {
            // restore primary states
            quickLightningEnabled.val = store_quickLightningEnabled;
            strobeEnabled.val = store_strobeEnabled;
            colorTweenEnabled.val = store_colorTweenEnabled;
        }

        private void SaveStateAll()
        {
            SaveState();
            CallActionAllQuickLightning("SaveState", false);
        }

        private void RestoreStateAll()
        {
            RestoreState();
            CallActionAllQuickLightning("RestoreState", false);
        }

        private void RestoreBaseIntensity()
        {
            JSONStorable light = GetLightStorable(containingAtom);
            if (light != null)
            {
                light.SetFloatParamValue("intensity", baseIntensity.val);
                previewCurrentIntensity.val = baseIntensity.val;
            }
        }

        private void RestoreBaseColor()
        {
            JSONStorable light = GetLightStorable(containingAtom);
            if (light != null)
            {
                light.SetColorParamValue("color", baseColor.val);
                previewCurrentColor.val = baseColor.val;
            }
        }

        /// <summary>
        /// Calls a specific QuickLightning ActionCallback on all lights in the scene.
        /// </summary>
        /// <param name="callAction"></param>
        /// <param name="backupAction"></param>
        private void CallActionAllQuickLightning(string callAction, bool turnOffNonQuickLightningLights)
        {
            foreach (Atom atom in SuperController.singleton.GetAtoms().Where((x) => { return (x != null && x.uid != containingAtom.uid && GetLightStorable(x) != null); }))
            {
                if (!CallActionQuickLightning(atom, callAction) && turnOffNonQuickLightningLights)
                {
                    JSONStorable light = GetLightStorable(atom);
                    if (light != null)
                    {
                        light.SetBoolParamValue("on", false);
                    }
                }
            }
        }

        /// <summary>
        /// Checks if the target atom has the same plugin as this one and calls action TurnOff to stop any ongoing actions and turns the lightning off.
        /// </summary>
        public static bool CallActionQuickLightning(Atom targetAtom, string callAction)
        {
            // KeyValuePair of Plugin -> Action (Action in plugin registered with RegisterAction(...) with a JSONStorableAction with the listed ID)
            List<KeyValuePair<string, string>> targetPluginActions = new List<KeyValuePair<string, string>>();
            targetPluginActions.Add(new KeyValuePair<string, string>("Blazedust.QuickLightning", callAction));
            
            bool pluginCalled = false;
            foreach (string receiverID in targetAtom.GetStorableIDs())
            {
                if (receiverID.StartsWith("plugin#"))
                {
                    foreach (KeyValuePair<string, string> pluginAction in targetPluginActions.Where((x) => { return receiverID.EndsWith(x.Key); }))
                    {
                        // Potential plugin!
                        // TODO: Check if plugin is enabled. For now assume plugin is enabled.
                        JSONStorable receiver = targetAtom.GetStorableByID(receiverID);
                        if (receiver != null)
                        {
                            foreach (string actionParam in receiver.GetActionNames())
                            {
                                if (pluginAction.Value == actionParam)
                                {
                                    receiver.CallAction(actionParam);
                                    pluginCalled = true;
                                }
                            }
                        }
                    }
                }
            }

            return pluginCalled;
        }

    }
}